#!/bin/bash

##########################################
# This script requires an OpenLDAP 2.5.x #
# with the overlay pw-totp and autoca    #
# active and configured!                 #
##########################################

# This script generats a TOTP-secretkey and a argon2 password for a user
# the script needs the full DN of the user
# and an email address for the user.
# The script will NOT check if the email address is valide
# Ther will be a random salt-string for the argon2-password
# and a string for the sharedkey generated for the user
# the user do not need the strings. The strings are only
# to generate the password and the sharedkey for the user.
# At the end you will see the sharedkey, the user can enter 
# in an authenticator app like googleauthentiicator or freeOTP
# and a png-file will be created with the QR-code for the sharedkey
#
# The TOTP needs some parameters, this paramters are stored in an OU.
# So you need at least one OU with the ObjectClass oathTOTPParams
#
# The second part of the script will generate a client-certificat 
# for the user. 
# This will only work, if the overlay "autca" is installed and active
#
# The script can be used via network or local only via the ldapi-socket

if [ $# -ne 2 ]
then
	echo "User-DN and User-Mail is needed"
	echo "usage: $0 userdn user@mail"
	exit 1
fi

USER_DN=$1
USER_MAIL=$2
# Default OU for TOTP parameter
# This is the OU with the ObjectClass oathTOTPParams
USER_OU="ou=users,dc=example,dc=net"
LDAP_SERVER=ldap://ldap25-p01.example.net

# Default Login to the LDAP-Server is via sasl-mech EXTERNAL 
# And the socket ldapi:///
USE_LDAPI=1 # set to "0" if userlogin is prefered
USE_TLS=1 # if TLS should not be used set to "0"
#LDAP_ADMIN="uid=ldap-admin,ou=users,dc=example,dc=net"
LDAP_ADMIN="cn=admin,dc=example,dc=net"
LDAP_ADMIN_PW="secret"
TOPT_ISSUER=stka
USER_NAME=""
# You will find all created files in this directory
PATH_FOR_SHARED_KEY="/root/"
USER_SHARED_KEY=""
QR_TEXT=""
USER_PASSWORD=""
# ARGON2_MEM 2^N KiB default is 12 (us as much as possible)
ARGON2_MEMORY=12
# ARGON2_ITERATION default is 3 (choos so that login takes not longer then 1 second)
ARGON2_ITERATION=3
# ARGON2_PARALISSM default is 1 (should be No. of CPU-cores * 2)
ARGON2_PARALISSM=1
# SALT_LENGTH should be at least 20 Bytes long
SALT_LENGTH=20

# Setting the variables to use TLS if selected
if [ "$USE_TLS" -eq 1 ]
then
	ACTIVATE_TLS="-ZZ"
else
	ACTIVATE_TLS=""
fi	

if [ ! -f "/usr/bin/qrencode" ]
then
	echo "qrencode is not installed!"
	exit 2
fi

if [ ! -f "/usr/bin/argon2" ]
then
	echo "argon2 is not installed!"
	exit 3
fi

#First part of DN is USER_NANE
USER_NAME=$(echo "$USER_DN" | cut -d "," -f 1)

# Let's get shure that there is exactly one user present in the database
if [ $USE_LDAPI -eq 1 ]
then
	USER_DN_IN_DB=$(ldapsearch -Q -Y EXTERNAL -LLL -H ldapi:///  $USER_NAME dn 2>/dev/null | cut -d " " -f 2)
else
	USER_DN_IN_DB=$(ldapsearch -xLLL "$ACTIVATE_TLS" -D "$LDAP_ADMIN" -w "$LDAP_ADMIN_PW" -H "$LDAP_SERVER" "$USER_NAME" dn 2>/dev/null | cut -d " " -f 2)
fi

if [ "${USER_DN,,}" != "${USER_DN_IN_DB,,}" ]
then
	echo "User $USER_DN not found in database"
	exit 4
fi	

# Look for the OU with TOTP-parameters
USER_OU=${USER_OU,,}

while [ "$CHANGE" != "y"  -a  "$CHANGE" != "n" ]
do
	echo "Default OU with TOTP-parameters is $USER_OU"
	echo ""
	echo -n "Change it? (y/n) "
	read CHANGE
	CHANGE=${CHANGE,,}
done

if [ $CHANGE == "y" ]
then
	while [ -z "$NEW_OU" ]
	do	
		echo "Enter DN of OU with TOTP-parameters : "
		read NEW_OU
		NEW_OU=${NEW_OU,,}
	done
	NEW_OU_CN=$(echo "$NEW_OU" | cut -d "," -f 1)
	if [ $USE_LDAPI -eq 1 ]
	then
        	NEW_OU_IN_DB=$(ldapsearch -Q -Y EXTERNAL -LLL -H ldapi:///  "$NEW_OU_CN" dn 2>/dev/null | cut -d " " -f 2)
	else
        	NEW_OU_IN_DB=$(ldapsearch -xLLL "$ACTIVATE_TLS" -D "$LDAP_ADMIN" -w "$LDAP_ADMIN_PW" -H "$LDAP_SERVER" "$NEW_OU_CN" dn 2>/dev/null | cut -d " " -f 2)
	fi
	NEW_OU_IN_DB=${NEW_OU_IN_DB,,}
	if [ "$NEW_OU" != "$NEW_OU_IN_DB" ]
	then
		echo "$NEW_OU not found "
		exit 5
	fi
	if [ $USE_LDAPI -eq 1 ]
	then
		PARMS=$(ldapsearch -Q -Y EXTERNAL -LLL -H ldapi:/// "$NEW_OU_CN" | grep "objectClass: oathTOTPParams")
	else
        	PARMS=$(ldapsearch -xLLL "$ACTIVATE_TLS" -D "$LDAP_ADMIN" -w "$LDAP_ADMIN_PW" -H "$LDAP_SERVER" "$NEW_OU_CN" | grep "objectClass: oathTOTPParams")
	fi
	if [ "$PARMS" != "objectClass: oathTOTPParams" -o -z "$PARMS" ]
	then
		echo "No TOTP-parameter found in $NEW_OU_IN_DB"
		exit 6
	fi
fi	

#Create random 20 byte string: 160 bit is recommended (min 128 bit)
# This will be the sharedkey for the user
USER_SHARED_KEY="${PATH_FOR_SHARED_KEY}${USER_NAME}-sharedkey.data"
touch $USER_SHARED_KEY
chmod 600 $USER_SHARED_KEY
openssl rand $SALT_LENGTH > $USER_SHARED_KEY
USER_SHARED_KEY_BASE32=$(base32 $USER_SHARED_KEY)

# Create the salt for the argon2 password
USER_SALT=$(tr -dc A-Za-z0-9 </dev/urandom | head -c 20 ; echo '')

#Create the password for 2fa
PW1=""
PW2=""
PW_CHECK=0
while [ "$PW_CHECK" -eq 0 ]
do
	while [ -z "$PW1" ]
	do
		echo -n "first run Password : " ;stty -echo ; read PW1 ; stty echo ; echo "";
	done
	while [ -z "$PW2" ]
	do
		echo -n "second run Password : " ;stty -echo ; read PW2 ; stty echo ; echo "";
	done
	if [ "$PW1" != "$PW2" ]
	then
		echo "Passwords do not match!"
	fi
	if [ "$PW1" == "$PW2" ]
	then
		PW_CHECK=1
	else
		PW1=""
		PW2=""	
	fi	
done
USER_PASSWORD=$PW1

ARGON_PASSWORD=$(echo -n "$USER_PASSWORD" | argon2 "$USER_SALT" -m "$ARGON2_MEMORY" -t "$ARGON2_ITERATION" -p "$ARGON2_PARALISSM" -e)

#Set user password
if [ $USE_LDAPI -eq 1 ]
then
	ldappasswd -Q -Y EXTERNAL -H ldapi:/// -s "$ARGON_PASSWORD" "$USER_DN"
else
	ldappasswd -x "$ACTIVATE_TLS" -D "$LDAP_ADMIN" -w "$LDAP_ADMIN_PW" -s "$ARGON_PASSWORD" "$USER_DN"
fi

# Create the LDIF-file to modify User for TOTP
while read LINE
do
	echo $LINE >> $PATH_FOR_SHARED_KEY/${USER_NAME}-totp.ldif
done << EOT	
dn: $USER_DN
changetype: modify
add: objectClass
objectClass: oathTOTPToken
-
add: oathTOTPParams
oathTOTPParams: $USER_OU
-
add: oathSecret
oathSecret:< file://$USER_SHARED_KEY
-
add: objectClass
objectClass: oathTOTPUser
-
add: oathTOTPToken
oathTOTPToken: $USER_DN
EOT



# Let's add TOTP ObjectClass and Attributes to the user
if [ $USE_LDAPI -eq 1 ]
then
	ldapmodify -Q -Y EXTERNAL -H ldapi:///  -f $PATH_FOR_SHARED_KEY/${USER_NAME}-totp.ldif
else
	ldapmodify -x "$ACTIVATE_TLS" -D "$LDAP_ADMIN" -w "$LDAP_ADMIN_PW" -H "$LDAP_SERVER" -f $PATH_FOR_SHARED_KEY/${USER_NAME}-totp.ldif
fi


#Create the QR-text for the user
QR_TEXT="otpauth://totp/$LDAP_SERVER:$USER_MAIL?secret=${USER_SHARED_KEY_BASE32}&issuer=${TOTP_ISSUER}&period=30&digits=6&algorithm=SHA1"

#Generate the QR-code
echo $QR_TEXT |  qrencode -s9 -o ${USER_MAIL}.png

echo "Shared key for User $USER_DN is $USER_SHARED_KEY_BASE32"
echo "You will find the key in data-format in file $USER_SHARED_KEY"
echo "You find the QR-code in file ${USER_MAIL}.png"

# Creating the users client-certificate
# only works if the overlay autoca is configured
echo ""
echo "Do you want to create the client certificate for the user too?"
echo "Only works if overlay autoca is configured in the main database"
while [  -z "$MAKE_CERT" ]
do
	echo -n "configure client certificate (y/n)? "
	read MAKE_CERT
	MAKE_CERT=$(echo ${MAKE_CERT^^})
done

if [ "$MAKE_CERT" == "Y" ]
then
	if [ $USE_LDAPI -eq 1 ]
	then	
		ldapsearch -Q -Y EXTERNAL -LLL -H ldapi:///  "$USER_NAME" "userCertificate;binary" "userPrivateKey;binary"
	else	
		ldapsearch -x -LLL "$ACTIVATE_TLS" -D "$LDAP_ADMIN" -w "$LDAP_ADMIN_PW" -H "$LDAP_SERVER" "$USER_NAME" "userCertificate;binary" "userPrivateKey;binary"
	fi	
else
	echo "No client certificate created to creat it do:"
	echo ldapsearch -x -LLL -D "$LDAP_ADMIN" -w LDAP_ADMIN_PW "$USER_NAME" "userCertificate;binary" "userPrivateKey;binary"
	echo "or"
	echo ldapsearch -Y EXTERNAL -H ldapi:/// -LLL  "$USER_NAME" "userCertificate\\;binary" "userPrivateKey\\;binary"
fi
